#The following code applies a portion of the MAFALDA procedure to data from the "realistic" time period from the parcels simulation. 
#The code then saves the outputs from this that we will use to make figure 4 

#Load modules
import datetime as dt
from numpy import dtype
import numpy as np
from pylab import *
import netCDF4 as n4
from netCDF4 import Dataset
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib import gridspec
import scipy.fftpack
from math import *
import xarray as xr
from scipy import integrate
from scipy import signal
from scipy.interpolate import RegularGridInterpolator
from scipy import spatial
from descartes import PolygonPatch
import shapely
from scipy.fft import fft, ifft
from shapely.geometry import Polygon
from shapely.geometry import LineString, Point
from PIL import Image
import os
import numpy.polynomial.polynomial as poly
from sklearn.metrics import r2_score
import numba
import matplotlib.ticker as ticker

## Define functions

#Execute the MAFALDA process using data from a given time period derived from the simulations.
def MAFALDA_process(SST, L_out, cval):
#INPUTS:
#L_out = the outer radius value, set to 600km for our study
#cval = the value of the constant, c, that is used to separate the eyewall and rainband circulations for their separate analyses.
#Specifically, if a trajectory derived from the isentropic streamfunction in this period experiences a maximum temperature 
#differential that exceeds a value of c * (SST-T_tropopause), then all trajectories at the corresponding level of the isentropic
#streamfunction are denoted as belonging to the eyewall circulation. If this is not the case, they belong to the rainband 
#circulation.
###############################################################################################################################
#This function has no outputs, but saves several variables. The saved variables are as follows:
#Two variables that are used by Parcel_plots.py to make the Figure 4 in the main text.
###############################################################################################################################
    #Define constants
    R = 287 #Gas constant for dry air (J/(K kg)). Matches the value in CM1
    Rv = 461.5 #Water vapor gas constant (J/(K kg)). Matches the value in CM1
    cpd = 1005.7 #Specific heat capacity of dry air in J/kg K. Matches the value in CM1
    Lv0 = 2501000 #Latent heat of vaporization of water in J/kg. Matches the value in CM1 
    kappa=R/cpd 
    T_f=273.15 #Freezing temperature (K)
    epsil=R/Rv
    p_ref=100000 #Reference pressure in Pa. Matches the value in CM1
    T_ref=T_f #Reference temperature in K. Matches the value in CM1

    #Load data
    #Specify the location of the files- the code below accomodates the specific folder-naming system that I used for the 
    #simulations.
    root= '/scratch/groups/oneillm/Laurel/SST_exps_3D/axisymm_exps/Parcel_tests/2-day-period/' 

    tstart=602 #The timestep at the beginning of the period
    tfinal=626 #The first timestep after the end of the period
    #Set the inner radius of the domain and optionally determine the radial index that corresponds to the input outer radius
    fileloc = root + 'cm1out_000602.nc' #Use the first data file from the simulation
    ds1=xr.open_dataset(fileloc)
    x_full=np.squeeze(ds1['xh']*1000).values #Full set of radial gridpoints in m
    L_o_ind=min(np.where(x_full>L_out*1000)[0]) #Index corresponding to the input outer radius
    z1=20 #Desired height in km up to which to load data
    z0=0 #Minimum height in km for loaded data
    L_outer=x_full[L_o_ind-1] #Outer radius. Same as input L_out, but in m.
    #print("Louter is "+np.str(L_outer))   
    L_inner=0
    L_i_ind=0
    
    #Load data from the axisymmetric CM1 simulations. The code assumes we saved the data in netcdf files
    x0 = L_inner; x1 = L_out; #Smallest and largest radii (km) for the loaded data.
    for ind in np.arange(tstart, tfinal): #Step through data files, startung at the timestep specified by "tstart"
        if (ind < 1000)&(ind >= 100): 
            fileloc = root + 'cm1out_000'+np.str(ind)+'.nc'
        elif (ind < 100)&(ind >= 10): 
            fileloc = root + 'cm1out_0000'+np.str(ind)+'.nc'     
        elif ind<10:
            fileloc = root + 'cm1out_00000'+np.str(ind)+'.nc'        
        else:
            fileloc = root + 'cm1out_00'+np.str(ind)+'.nc'
        ds=xr.open_dataset(fileloc) #Open the netcdf file corresponding to the current timestep
        if ind==tstart: #Initialize all data arrays for the special case of ind=tstart
            temp_p=np.squeeze(ds['prs'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values #Pressure (Pa)
            temp_p=temp_p[..., np.newaxis]
            p=temp_p
            temp_th=np.squeeze(ds['th'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values #Potential temperature (K)
            temp_th=temp_th[..., np.newaxis]
            th=temp_th
            temp_qv=np.squeeze(ds['qv'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values #Water vapor mixing ratio (kg/kg)
            temp_qv=temp_qv[..., np.newaxis]
            qv=temp_qv
            temp_w=np.squeeze(ds['winterp'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values #Vertical wind speed (m/s)
            temp_w=temp_w[..., np.newaxis]
            w=temp_w
            temp_rho=np.squeeze(ds['rho'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values #Dry-air density (kg/m^3)
            temp_rho=temp_rho[..., np.newaxis]
            rho=temp_rho
            x=np.squeeze(ds['xh'].sel(xh=slice(x0,x1))*1000).values #Radial grid point values. Convert to m
            dx=np.mean(np.diff(x))  
            z=np.squeeze(ds['zh'].sel(zh=slice(z0,z1))*1000).values #Vertical grid points. Convert to m. 
            time=np.squeeze(ds['time'])
        #For the cases where the output variables have already been initialized, append data from the current timestamp to 
        #the existing arrays 
        else:
            temp_rho=np.squeeze(ds['rho'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values
            temp_rho=temp_rho[..., np.newaxis]
            rho=np.concatenate((rho, temp_rho), axis=2) 
        
            temp_w=np.squeeze(ds['winterp'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values
            temp_w=temp_w[..., np.newaxis]
            w=np.concatenate((w, temp_w), axis=2) 
        
            temp_th=np.squeeze(ds['th'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values
            temp_th=temp_th[..., np.newaxis]
            th=np.concatenate((th, temp_th), axis=2) 
        
            temp_qv=np.squeeze(ds['qv'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values
            temp_qv=temp_qv[..., np.newaxis]
            qv=np.concatenate((qv, temp_qv), axis=2)
        
            time=np.squeeze(ds['time']) 
        
            temp_p=np.squeeze(ds['prs'].sel(zh=slice(z0,z1),xh=slice(x0,x1))).values
            temp_p=temp_p[..., np.newaxis]
            p=np.concatenate((p, temp_p), axis=2) 

    r=x #Radial gridpoints
            
    #Calculate temperature and equivalent potential temperature
    t=th*(p/p_ref)**kappa #Temperature (K)
    #Find equivalent potential temperature. Matches the formulation in CM1.
    ee=0.01*p*qv/(epsil+qv) #Same as ep but in hPa. Matches CM1 formulation.
    tlcl=55.0+2840.0/(3.5*np.log(t)-np.log(1.0e-20+ee)-4.805) #Temperature at the LCL. Matches CM1 formulation.   
    th_e=t*((p_ref/p)**(0.2854*(1.0-0.28*qv)))*np.exp(((3376.0/tlcl)-2.54)*qv*(1.0+0.81*qv)) #Equivalent potential temperature. Matches CM1 formulation

    #Find the anomalous (with respect to the horizontal average) vertical mass flux, rho*w. 
    f=rho*w #Vertical mass flux
    #Remove horizontal-mean vertical mass flux from the vertical mass flux to get the anomalous vertical mass flux
    perArea=np.sum(x) 
    xlen=x.size
    zlen=z.size
    x=x.reshape(1, xlen, 1) #Change the shape of x for the following calculation
    f_av= np.sum(f*x, axis=1)/perArea #Horizontal mean of the vertical mass flux. This has the wrong dimensions for what we want to do next, so in what follows we will pad 
                                      #it such that it has the same dimensions as f.
    f_avg=np.zeros((zlen, (L_o_ind-L_i_ind), (tfinal-tstart)))  #Initialize f_avg, which will serve as the "padded" version of f_av, with the same dimensions as f
    for xind in np.arange(0, (L_o_ind-L_i_ind)):
        f_avg[:, int(xind), :]=f_av #For each radius, set a column of f_avg to be equal to f_av. This ensures f_avg is equivalent to f_av but with the same dimensions as f
    f=f-f_avg #Remove horizontal-mean vertical mass flux from f to get the anomalous vertical mass flux
    x=x.reshape(xlen) #Change the shape of x back to what it was before


    #Perform the isentropic integral of the vertical mass flux, based on a modified verion of Mrowiec et al. (2016, JAS) equation (2).
    #We approximate the theta_e delta function in this formula by binning the mass flux at all according to the value of 
    #the equivalent potential temperature theta_e.

    #Set parameters of the theta_e bins.
    delt=0.5 #Half-width of theta_e bins
    th_min=295 #Minimum value of theta_e for the bins
    th_max=450 #Maximum value of theta_e for the bins

    number=(th_max-th_min)/(2*delt) #Number of theta_e bins
    bins=np.arange(th_min-delt, th_max+delt, delt*2) #Define theta e bins based on the parameters th_min, th_max, and delt

    #Perform the isentropic integral to get <rho*w>(z, theta_e, t), and store in f2. We then divide by perArea to get the isentropic-average vertical mass flux
    #The following code loops through theta_e bins defined above and searches for the points (r, z, t) with theta_e values that fall inside these bins
    #Once these points (r, z, t) are obtained, the code performs the isentropic horizontal integral for each z value. The result of this is stored in f2b.
    #After this, the vector f2b (which has dimensions zx1xt) is added to the array f2 as a column. f2 has dimensions of zx(number of theta_e bins)xt
    for b in bins: #"For a given theta e bin, do the following"
        if b != bins[int(number)]: #Don't deal with last b value since it is redundant given how I write the next line- would be like counting the last bin twice
            [zindex, xindex, tindex]=np.where((th_e>b) & (th_e<=b+1)) #Get indices of all points (r, z, t) at which theta_e values fall within the theta_e bin of interest. 
            #In what follows, a single such point (r, z, t) would be given by (x[xindex[i]], z[zindex[i]], t[tindex[i]]), where i is some index between 0 and the length of 
            #xindex.
            f2b=np.zeros((zlen, 1, (tfinal-tstart))) #Initialize f2b, which will hold the result of the isentropic integral for given values of z, t, and a given theta e 
                                                     #bin. Dimensions are zx1xt
            #Now that we have found the points (r, z, t) of interest for a given theta_e bin, we calculate the isentropic integral wrt r for each z value
            for i in np.unique(zindex): #"Do the following for a given z value in the set of points denoted by zindex, xindex, and tindex"
                for m in np.unique(tindex[zindex==i]): #"Within the subset of t values corresponding to the current z value, do the following"
                    integrand=[] 
                    indices=np.where((zindex==i)&(tindex==m))[0] #For the chosen z and t values, z' and t', find all points (r, z=z', t=t')
                    #For the chosen z and t values, z' and t', calculate the integrand at points (r, z=z', t=t'): (rho*w')(r, z=z', t=t')*r 
                    for j in indices:
                        integrand.append(f[i, xindex[j], m]*x[xindex[j]]) #Calculate the integrand, (rho*w)'(r, z=z', t=t')*r , at all points (r, z=z', t=t')    
                    f2b[i, :, m]=np.sum(integrand, axis=0) #Perform the isentropic integral: <rho*w>(theta_e bin, z=z', t=t')=integral((rho*w)'(r, z=z', t=t')*r dr, limits:
                                                           #(0, L_outer))
            #f2b contains the result of the isentropic integral for all values of t and z corresponding to the current theta_e bin.
            #Now, we add f2b to f2 in the position corresponding to the current theta_e bin.
            if b==bins[0]: #For the first bin, we need to handle things as a special case to initialize f2
                    f2=f2b #Initialize f2, which contains the result of the isentropic integral at all values of z, t, and theta_e
            else: 
                f2=np.concatenate((f2, f2b), axis=1) #Concatenate f2b along the theta_e dimension of f2
    f2=f2/perArea #Divide f2 by the horizontal area of the domain to obtain the isentropic-average vertical mass flux, <rho*w>(z, theta_e, t)       

    #Time integral of f2.
    f_tavg=np.sum(f2, axis=2) #Integrate along the time axis. No need to multiply by dt because if we divide by tfinal-tstart (time period in units of # of timestamps) then
                              #dt=1 timestamp
    f_tavg=f_tavg/(tfinal-tstart) #Divide bu tau=tfinal-tstart


    #Find the isentropic streamfunction using the non-time-averaged isentropic-average vertical mass flux, following a modified version of equation (4) of Mrowiec et al. 
    #(2016, JAS)
    [zlen, thelen]=shape(f_tavg)
    thevec=arange(thelen) #Indices in theta_e dimension
    for i in thevec:
        if i==0:
            f3=f2[:, i, :]*2*delt #Perform integral, psi(z, theta_e)=integral(<rho*w>(z, theta_e', t) dtheta_e', limits: (0, theta_e)), for the first theta_e value
            f3=f3[..., np.newaxis] #Gives f3 the correct shape, so we can append it to psi_t
            psi_t=f3 #Add the result of the integral to psi_t
        else:
            f3=np.sum((f2[:, 0:i, :]*2*delt), axis=1) #Perform integral, psi(z, theta_e)=integral(<rho*w>(z, theta_e', t) dtheta_e', limits: (0, theta_e)), for a given 
                                                      #theta_e value
            f3=f3[..., np.newaxis] #Gives f3 the correct shape, so we can append it to psi_t
            psi_t=np.concatenate((psi_t, f3), axis=2) #Add the result of the integral to psi_t 
    psi=np.sum(psi_t, axis=1) #Find time-average isentropic streamfunction. As before, no need to multiply by dt because if we divide by tfinal-tstart (time period in units
                              #of # of timestampe) then dt=1 timestamp
    psi=psi/(tfinal-tstart) #Divide bu tau=tfinal-tstart    

    ##OPTIONAL- Plot isentropic streamfunction. 
    #plt.rc('axes', labelsize=40)    # fontsize of the x and y labels
    #plt.rc('xtick', labelsize=40)    # fontsize of the x tick labels
    #plt.rc('ytick', labelsize=40)    # fontsize of the x tick labels
    #plt.rc('legend', fontsize=40)            # fontsize of the legend text
    #plt.figure(figsize=(15,10)) 
    theta_es=np.arange(th_min, th_max, delt*2)
    [zplot, thplot]=np.meshgrid(theta_es, z)
    #ncontour = 25
    #contourarray = psi.min()*((2*np.arange((ncontour+1),1.0, -1.0)-1)/(2*ncontour))
    #plt.contour(zplot, thplot/1000, psi,contourarray, cmap='rainbow') #Set to 25 contours not 30 bc this is what we use later!
    #plt.colorbar(format=ticker.FuncFormatter(fmt))
    #plt.title('Isentropic Streamfunction (kg $m^{-2}$ $s^{-1}$)', fontsize=30)
    #plt.ylabel('Height (km)')
    #plt.xlabel('Equivalent Potential Temperature (K)')
    #xlim(th_min, th_max)

    #Get the coordinates of all levels of the isentropic streamfunction from a contour plot of it. 
    #We will use the closed loops at each level of the streamfunction to construct "MAFALDA trajectories".
    #These trajectories are assumed to represent the mean trajectories of real air parcels in the TC in z-theta_e space.
    plt.figure() #Make a new figure so we don't overplot on the previous one 
    ncontour = 25 #Number of levels to use in our contour plot of the isentropic streamfunction
    contourarray = psi.min()*((2*np.arange((ncontour+1),1.0, -1.0)-1)/(2*ncontour)) #Sets the levels of the isentropic streamfunction that we will plot as contours
    h1 = plt.contour(zplot,thplot,psi,contourarray,cmap='rainbow') #Contour plot of the isentropic streamfunction
    contcoords = h1.allsegs #Coordinates of MAFALDA trajectories corresponding to each level of the isentropic streamfunction
    contlevels=h1.levels #Values of all levels of the isentropic streamfunction

    
    #Interpolate the mass-weighted isentropic- and time-averaged variables onto the gridpoints of the MAFALDA trajectories.
    interpolant_th=[]
    interpolant_z=[]
    #Assign maximum theta_e and z values for each SST, where a MAFALDA trajectory that exceeds these values is ignored by the code.
    #This code is included because the isentropic-averaging procedure sometimes (rarely) produces "contrail" artefacts- mostly-horizontal lines
    #in z-theta_e space that extend upwards in theta_e from the isentropic streamfunction. These "contrails" are artefacts of the code 
    #and are not physical, so we ignore them in our analysis.
    #These maximum values are set by inspection of the isentropic streamfunctions.
    maxt=370
    maxz=15000
    for cont in np.arange(0, len(contcoords)): #Loop through the levels of the isentropic streamfunction. 
        clevel=contlevels[cont]
        if contcoords[cont]: #Only do something if the list for the chosen contour is not empty
            for j in np.arange(0, len(contcoords[cont])): #Loop through all MAFALDA trajectories at this level of the isentropic streamfunction. There may be more than one 
                                                          #of these.
                ccont=contcoords[cont][j]
                if (max(ccont[:, 0])<maxt)&(max(ccont[:, 1]<maxz)): #Don't count "contrails"
                    interpolant_th_ar=ccont[:, 0] #Theta_e points on the trajectory
                    interpolant_z_ar=ccont[:, 1] #z points on the trajectory
                    interpolant_th_l=interpolant_th_ar.tolist()
                    interpolant_z_l=interpolant_z_ar.tolist()
                    if (interpolant_th_l[0]!=interpolant_th_l[-1]): #To construct a trajectory that is truly a "closed loop", we must add the first point in the 
                                                                    #trajectory as its last point.
                        interpolant_th_l.append(interpolant_th_l[0])
                        interpolant_z_l.append(interpolant_z_l[0])
                    interpolant_th_a=np.asarray(interpolant_th_l)
                    interpolant_z_a=np.asarray(interpolant_z_l)
                    interpolant_th.append(interpolant_th_a) #Contains the sets of theta_e points for all MAFALDA trajectories
                    interpolant_z.append(interpolant_z_a) #Contains the sets of z points for all MAFALDA trajectories
    interpolant_th_arr=np.asarray(interpolant_th) #Convert to a numpy array
    interpolant_z_arr=np.asarray(interpolant_z)

    np.savez("Interpolated_trajs.npz", interpolant_th_arr=interpolant_th_arr, interpolant_z_arr=interpolant_z_arr)





#The following code is for formatting of plot labels only. It converts labels to scientific notation. Source: 
#https://python.tutorialink.com/scientific-notation-colorbar-in-matplotlib/
def fmt(x, pos): 
    a, b = '{:.2e}'.format(x).split('e')
    b = int(b)
    return r'${} \times 10^{{{}}}$'.format(a, b)


#The following code applies the MAFALDA procedure to the 500 parcels simulation results
SST=299.5
L_out=600 #Outer radius (km) for the MAFALDA analysis. This value is subjectively set based on the single-eyewall period data from the main simulations.
cval=0.82 #Eyewall circulation threshold (dimensionless). Minimum fraction of (SST-T_tropopause) temperature difference 
          #represented by "eyewall trajectories". This threshold is subjectively set based on the single-eyewall period data from the main simulations.
MAFALDA_process(SST, L_out, cval) #Perform the MAFALDA procedure on the data from the selected realistic period.
        
